#include "SDL.h"
#include "SDL_Mixer.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "MapInfo.h"
#include "Hero.h"
#include "Exit.h"
#include "Enemy.h"

const int MAX_ENEMIES = 100;
const int MIN_ENEMIES = 7;

const Uint32 MIN_TICKS = 10;

const int GAME_NEW_LEVEL = 0;
const int GAME_PAUSE = 1;
const int GAME_PLAY = 2;
const int GAME_OVER = 3;
const int GAME_TITLE = 4;
const int GAME_LEVEL_COMPLETE = 5;

bool isCollision(Sprite* sprA, Sprite* sprB)
{

	const int ERROR_X = 10;
	const int ERROR_Y = 8;

	int dx;
	int dy;

	int minWidth;
	int minHeight;

	Uint32 distA;
	Uint32 distB;

	dx = abs((sprA->getX() + (sprA->getWidth() / 2)) -
		     (sprB->getX() + (sprB->getWidth() / 2)));
	dy = abs((sprA->getY() + (sprA->getHeight() / 2)) -
		     (sprB->getY() + (sprB->getHeight() / 2)));

	minWidth = sprA->getWidth();
	minHeight = sprA->getHeight();

	if (sprB->getWidth() < minWidth)
		minWidth = sprB->getWidth();

	if (sprB->getHeight() < minHeight)
		minHeight = sprB->getHeight();

	if ((dx > minWidth) || (dy > minHeight))
	{
		return 0;
	}
	else
	{
		//Shrink the circle a bit to make sure it is a hit.
		//if (minWidth > ERROR_X)
		//	minWidth -= ERROR_X;

		//if (minHeight > ERROR_Y)
		//	minHeight -= ERROR_Y;

		minWidth /= 2;
		minHeight /= 2;

		distA = (minWidth * minWidth) + (minHeight * minHeight);
		distB = (dx * dx) + (dy * dy);

		if (distB < distA)
		{
			return true;
		}
		else
		{
			return (distB < distA);
		}
	}
}

void randomWalk(int* maze, int centerX, int centerY, bool bUpDown, int callsLeft, int mapWidth, int mapHeight)
{

	int i;
	int j;
	int iSteps1;
	int iSteps2;

	if (callsLeft > 0)
	{
		iSteps1 = ((rand() % 3) * 2) + 2;
		iSteps2 = ((rand() % 3) * 2) + 2;

		if (bUpDown)
		{
			//Draw up and down until number of steps or a blank found
			if (centerY - iSteps1 <= 0)
			{
				iSteps1 = centerY;
			}
			if (centerY + iSteps2 >= mapHeight - 1)
			{
				iSteps2 = mapHeight - centerY - 1;
			}

			while ((iSteps1 % 2) != 0)
				iSteps1--;

			while ((iSteps2 % 2) != 0)
				iSteps2--;

			// Draw up
			for(j=centerY; j >= centerY - iSteps1; --j)
			{
				if ((maze[(j * mapWidth) + centerX] == 0) && (j != centerY))
				{
					j--;
					break;
				}
				else
				{
					maze[(j * mapWidth) + centerX] = 0;
				}
			}
			randomWalk(maze, centerX, j + 1, false, callsLeft - 1, mapWidth, mapHeight);

			// Draw down
			for(j=centerY; j <= centerY + iSteps2; ++j)
			{
				if ((maze[(j * mapWidth) + centerX] == 0) && (j != centerY))
				{
					j++;
					break;
				}
				else
				{
					maze[(j * mapWidth) + centerX] = 0;
				}
			}
			randomWalk(maze, centerX, j - 1, false, callsLeft - 1, mapWidth, mapHeight);

		}
		else
		{
			//Draw left and right until number of steps or blank found
			if (centerX - iSteps1 <= 0)
			{
				iSteps1 = centerX;
			}
			if (centerX + iSteps2 >= mapWidth - 1)
			{
				iSteps2 = mapWidth - centerX - 1;
			}

			// Draw Left
			for(i=centerX; i >= centerX - iSteps1; --i)
			{
				if ((maze[(centerY * mapWidth) + i] == 0) && (i != centerX))
				{
					i--;
					break;
				}
				else
				{
					maze[(centerY * mapWidth) + i] = 0;
				}
			}
			randomWalk(maze, i + 1, centerY, true, callsLeft - 1, mapWidth, mapHeight);

			// Draw down
			for(i=centerX; i <= centerX + iSteps2; ++i)
			{
				if ((maze[(centerY * mapWidth) + i] == 0) && (i != centerX))
				{
					i++;
					break;
				}
				else
				{
					maze[(centerY * mapWidth) + i] = 0;
				}
			}		
			randomWalk(maze, i - 1, centerY, true, callsLeft - 1, mapWidth, mapHeight);
		}
	}
}

void randomMaze(int* maze, int mapX, int mapY, int mapWidth, int mapHeight)
{
	//Draw find a point in the center and draw lines from left, right, top and bottom
	//calls to the line drawing functions will draw new perpendicular lines of
	//diminishing size.  When they are short enough, the recursive functions will stop

	const int NUM_ITERATIONS = 5;

	int centerX;
	int centerY;

	int i;
	int j;

	for (j=0;j<mapHeight;++j)
	{
		for(i=0;i<mapWidth;++i)
		{
			maze[(j*mapWidth) + i] = 1;
		}
	}

	// Map looks really bad if we don't start at the center.
	centerX = mapWidth / 2;
	centerY = mapHeight / 2;

	randomWalk(maze, centerX, centerY, false, NUM_ITERATIONS, mapWidth, mapHeight);
	randomWalk(maze, centerX, centerY, true, NUM_ITERATIONS, mapWidth, mapHeight);
}

int main(int argc, char** argv)
{

	const int SCREEN_WIDTH = 640;
	const int SCREEN_HEIGHT = 480;
	const int COLOR_DEPTH = 16;

	const int TILE_COUNT = 2;
	const int TILE_WIDTH = 32;
	const int TILE_HEIGHT = 32;

	const int MAP_WIDTH = (SCREEN_WIDTH / TILE_WIDTH);
	const int MAP_HEIGHT = (SCREEN_HEIGHT / TILE_HEIGHT);

	const char* TILES_FILE = "data\\tiles.bmp";

	bool bFullScreen = true;
	bool bQuit = false;

	Enemy* enemyList[MAX_ENEMIES];

	Exit* exitSprite = NULL;
	Hero* heroSprite = NULL;

	int* map;

	int i;
	int j;
	int newEnemyIDX = -1;
	int gameState = GAME_TITLE;
	int tileY = 0;
	
	Uint32 prevTime;
	Uint32 currTime;
	Sint32 enemyWait;

	MapInfo* mapInfo = NULL;

	Mix_Chunk* welcomeSound;
	Mix_Chunk* gameOverSound;
	Mix_Chunk* levelCompleteSound;
	
	SDL_Event event;
	SDL_Rect tileRect[TILE_COUNT];
	SDL_Rect writeRect;
	SDL_Surface* enemyGraph = NULL;
	SDL_Surface* exitGraph = NULL;
	SDL_Surface* gameOverGraph = NULL;
	SDL_Surface* heroGraph = NULL;
	SDL_Surface* levelCompleteGraph = NULL;
	SDL_Surface* pauseGraph = NULL;
	SDL_Surface* screen = NULL;
	SDL_Surface* temp = NULL;
	SDL_Surface* titleGraph = NULL;
	SDL_Surface* tiles = NULL;

	bFullScreen = true;
	for (i=0;i<argc;i++)
	{
		if (strcmp("win", argv[i])==0)
			bFullScreen = false;
	}

	/* Initialize the SDL library */
	if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO) < 0)
	{
		fprintf (stderr, "Could not initialize SDL:  %s\n", SDL_GetError());
		exit(1);
	}

	// Clean up on exit.
	atexit(SDL_Quit);

    // Initialize the display in a 640x480 16-bit mode,
    if (bFullScreen)
		screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, COLOR_DEPTH, SDL_HWSURFACE | SDL_DOUBLEBUF | SDL_FULLSCREEN); 
	else
		screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, COLOR_DEPTH, SDL_HWSURFACE | SDL_DOUBLEBUF); 

	if (screen == NULL)
	{
		fprintf(stderr, "Couldn't set %dx%x%d video mode: %s\n", SCREEN_WIDTH, SCREEN_HEIGHT, COLOR_DEPTH, SDL_GetError());
		exit(1);
	}

	tiles = SDL_LoadBMP(TILES_FILE);
	if (tiles == NULL)
	{
		fprintf(stderr, "Couldn't load %s: %s\n", TILES_FILE, SDL_GetError());
		exit(1);
	}

	titleGraph = SDL_LoadBMP("data\\title.bmp");
	if (titleGraph == NULL)
	{
		fprintf(stderr, "Couldn't load %s: %s\n", "data\\title.bmp", SDL_GetError());
		exit(1);
	}


	/* Load the file for the hero*/
    temp = SDL_LoadBMP("data\\hero_sprite.bmp");
    if (temp == NULL) {
        fprintf(stderr, "Couldn't load %s: %s\n", "data\\hero_sprite.bmp", SDL_GetError());
        return -1;
    }

    /* Set the transparent color */
	SDL_SetColorKey(temp,SDL_SRCCOLORKEY, SDL_MapRGB(temp->format,0,255,0));
	heroGraph = SDL_DisplayFormat(temp);
	SDL_FreeSurface(temp);
	temp = NULL;

	/* Load the file for the exit */
    temp = SDL_LoadBMP("data\\exit_sprite.bmp");
    if (temp == NULL) {
        fprintf(stderr, "Couldn't load %s: %s\n", "data\\exit_sprite.bmp", SDL_GetError());
        return -1;
    }

    /* Set the transparent color */
	SDL_SetColorKey(temp,SDL_SRCCOLORKEY, SDL_MapRGB(temp->format,0,255,0));
	exitGraph = SDL_DisplayFormat(temp);
	SDL_FreeSurface(temp);
	temp = NULL;
	
	/* Load the file for the enemy */
    temp = SDL_LoadBMP("data\\enemy_sprite.bmp");
    if (temp == NULL) {
        fprintf(stderr, "Couldn't load %s: %s\n", "data\\enemy_sprite.bmp", SDL_GetError());
        return -1;
    }

    /* Set the transparent color */
	SDL_SetColorKey(temp,SDL_SRCCOLORKEY, SDL_MapRGB(temp->format,0,255,0));
	enemyGraph = SDL_DisplayFormat(temp);
	SDL_FreeSurface(temp);
	temp = NULL;

	/* Load the file for the gameOver sign */
    temp = SDL_LoadBMP("data\\gameOver.bmp");
    if (temp == NULL) {
        fprintf(stderr, "Couldn't load %s: %s\n", "data\\gameOver.bmp", SDL_GetError());
        return -1;
    }

    /* Set the transparent color */
	SDL_SetColorKey(temp,SDL_SRCCOLORKEY, SDL_MapRGB(temp->format,255, 0, 255));
	gameOverGraph = SDL_DisplayFormat(temp);
	SDL_FreeSurface(temp);
	temp = NULL;

	/* Load the file for the pause sign */
    temp = SDL_LoadBMP("data\\pause.bmp");
    if (temp == NULL) {
        fprintf(stderr, "Couldn't load %s: %s\n", "data\\pause.bmp", SDL_GetError());
        return -1;
    }

    /* Set the transparent color */
	SDL_SetColorKey(temp,SDL_SRCCOLORKEY, SDL_MapRGB(temp->format,255, 0, 255));
	pauseGraph = SDL_DisplayFormat(temp);
	SDL_FreeSurface(temp);
	temp = NULL;

	/* Load the file for the levelComplete sign */
    temp = SDL_LoadBMP("data\\levelComplete.bmp");
    if (temp == NULL) {
        fprintf(stderr, "Couldn't load %s: %s\n", "data\\levelComplete.bmp", SDL_GetError());
        return -1;
    }

    /* Set the transparent color */
	SDL_SetColorKey(temp,SDL_SRCCOLORKEY, SDL_MapRGB(temp->format,255, 0, 255));
	levelCompleteGraph = SDL_DisplayFormat(temp);
	SDL_FreeSurface(temp);
	temp = NULL;

	//Start up SDL_Mixer
	if (Mix_OpenAudio(MIX_DEFAULT_FREQUENCY, MIX_DEFAULT_FORMAT, 1, 1024)==-1)
	{
		fprintf(stderr, "Couldn't start up SDL_Mixer:  %s\n", Mix_GetError());
		return -1;
	}

	welcomeSound = Mix_LoadWAV("data\\welcome.wav");
	if (!welcomeSound)
	{
		fprintf(stderr, "Couldn't load %s:  %s\n", "data\\welcome.wav", Mix_GetError());
		return -1;
	}

	gameOverSound = Mix_LoadWAV("data\\gameOver.wav");
	if (!gameOverSound)
	{
		fprintf(stderr, "Couldn't load %s:  %s\n", "data\\gameOver.wav", Mix_GetError());
		return -1;
	}

	levelCompleteSound = Mix_LoadWAV("data\\levelComplete.wav");
	if (!levelCompleteSound)
	{
		fprintf(stderr, "Couldn't load %s:  %s\n", "data\\gameOver.wav", Mix_GetError());
		return -1;
	}

	srand((unsigned)time(NULL));
	tileRect[0].x = 0;
	tileRect[0].y = 0;
	tileRect[0].w = TILE_WIDTH;
	tileRect[0].h = TILE_HEIGHT;
	tileRect[1].x = TILE_WIDTH - 1;
	tileRect[1].y = 0;
	tileRect[1].w = TILE_WIDTH;
	tileRect[1].h = TILE_HEIGHT;

	//Width and Height ignored but I am adding it in anyway.
	writeRect.x = 0;
	writeRect.y = 0;
	writeRect.w = TILE_WIDTH;
	writeRect.h = TILE_HEIGHT;

	map = new int[MAP_WIDTH * MAP_HEIGHT];

	for(i = 0;i<MAX_ENEMIES;++i)
	{
		enemyList[i] = NULL;
	}

	Mix_AllocateChannels(16);
	Mix_PlayChannel(-1,welcomeSound, 0); 

	while (!bQuit)
	{

		if (gameState == GAME_TITLE)
		{
			writeRect.x = 0;
			writeRect.y = 0;
			writeRect.w = titleGraph->w;
			writeRect.h = titleGraph->h;
			SDL_BlitSurface(titleGraph, NULL, screen, &writeRect);
			
			//Flip the surface
			SDL_Flip(screen);
		}
		else if (gameState == GAME_NEW_LEVEL)
		{

			tileY = (rand() % (tiles->h / TILE_HEIGHT)) * TILE_HEIGHT;
			for(i=0;i<TILE_COUNT;++i)
			{
				tileRect[i].y = tileY;
			}

			// delete old variables
			for(i=0;i<MAX_ENEMIES;++i)
			{
				if (enemyList[i] != NULL)
				{
					delete enemyList[i];
					enemyList[i] = NULL;
				}
			}

			if (exitSprite != NULL)
				delete exitSprite;

			if (heroSprite != NULL)			
				delete heroSprite;

			randomMaze(map, 0, 0,  MAP_WIDTH, MAP_HEIGHT);

			mapInfo = new MapInfo(map, MAP_WIDTH, MAP_HEIGHT);
			heroSprite = new Hero(heroGraph, mapInfo);

			//Find a spot for the hero to start at
			i = rand() % MAP_WIDTH;
			j = rand() % MAP_HEIGHT;
			while (map[(j*MAP_WIDTH) + i] != 0)
			{
				i = rand() % MAP_WIDTH;
				j = rand() % MAP_HEIGHT;
			}
			heroSprite->setX(i * 32);
			heroSprite->setY(j * 32); // Sorry, magic number.
			heroSprite->setActive(true);

			exitSprite = new Exit(exitGraph, mapInfo, heroSprite);
			exitSprite->setActive(true);

			for(i = 0;i<MAX_ENEMIES;++i)
			{
				enemyList[i] = NULL;
			}

			//Lets add the minimum number of enemies
			for(i=0; i< MIN_ENEMIES;++i)
			{
				enemyList[i] = new Enemy(enemyGraph, mapInfo, heroSprite);
				enemyList[i]->setActive(true);
			}
			// New enemy in 3-8 seconds.
			enemyWait = (rand() % 5000) + 3000;

			gameState = GAME_PLAY;

			currTime = SDL_GetTicks();
			prevTime = currTime - MIN_TICKS;
			if (prevTime < 0)
				prevTime = 0;

		}

		if ((gameState == GAME_PLAY) || (gameState == GAME_PAUSE) || 
			(gameState == GAME_OVER) || (gameState == GAME_LEVEL_COMPLETE))
		{
			currTime = SDL_GetTicks();
			if (currTime - prevTime >= MIN_TICKS)
			{
				if (gameState == GAME_PLAY)
				{				
					enemyWait -= abs(currTime - prevTime);
					prevTime = currTime;

					if (enemyWait <= 0)
					{
						// New enemy in 3-8 seconds.
						enemyWait = (rand() % 5000) + 3000;
						// Look for a new enemy to add
						newEnemyIDX = -1;
						for(i=0;i<MAX_ENEMIES;++i)
						{
							if (enemyList[i] == NULL)
							{
								newEnemyIDX = i;
								break;
							}
						}

						if (newEnemyIDX != -1)
						{
							enemyList[newEnemyIDX] = new Enemy(enemyGraph, mapInfo, heroSprite);
							enemyList[newEnemyIDX]->setActive(true);
						}
					}

					heroSprite->update();
					exitSprite->update();
					for(i=0;i<MAX_ENEMIES;++i)
					{
						if (enemyList[i]!=NULL)
						{
							enemyList[i]->update();
						}
					}

					// Check to see if the hero intersects with an exit or enemy
					// if enemy then game over.

					for(i = 0; i<MAX_ENEMIES;++i)
					{
						if (enemyList[i]!=NULL)
						{
							if (isCollision(enemyList[i], heroSprite))
							{
								gameState = GAME_OVER;
								Mix_PlayChannel(-1, gameOverSound, 0); 
								break;
							}
						}
					}

					if (gameState != GAME_OVER)
					{
						if (isCollision(exitSprite, heroSprite))
						{
							gameState = GAME_LEVEL_COMPLETE;
							Mix_PlayChannel(-1, levelCompleteSound, 0); 
						}
					}

				}

				// Draw to screen
				writeRect.y = 0;

				for (j=0;j<MAP_HEIGHT;++j)
				{
					writeRect.x = 0;
					for(i=0;i<MAP_WIDTH;++i)
					{
						if (SDL_BlitSurface(tiles, &tileRect[map[(j*MAP_WIDTH) + i]], screen, &writeRect) < 0)
						{
							fprintf(stderr, "BlitSurface Error:  %s\n", SDL_GetError());
						}
						writeRect.x += TILE_WIDTH;
					}
					writeRect.y += TILE_HEIGHT;
				}

				for(i=0;i<MAX_ENEMIES;++i)
				{
					if (enemyList[i]!=NULL)
					{
						enemyList[i]->Draw(screen);
					}
				}
				heroSprite->Draw(screen);
				exitSprite->Draw(screen);

				if (gameState == GAME_PAUSE)
				{
					writeRect.x = (screen->w - pauseGraph->w) / 2;
					writeRect.y = (screen->h - pauseGraph->h) / 2;
					writeRect.w = pauseGraph->w;
					writeRect.h = pauseGraph->h;
					SDL_BlitSurface(pauseGraph, NULL, screen, &writeRect);
				}
				else if (gameState == GAME_OVER)
				{
					writeRect.x = (screen->w - gameOverGraph->w) / 2;
					writeRect.y = (screen->h - gameOverGraph->h) / 2;
					writeRect.w = gameOverGraph->w;
					writeRect.h = gameOverGraph->h;
					SDL_BlitSurface(gameOverGraph, NULL, screen, &writeRect);
				}
				else if (gameState == GAME_LEVEL_COMPLETE)
				{
					writeRect.x = (screen->w - levelCompleteGraph->w) / 2;
					writeRect.y = (screen->h - levelCompleteGraph->h) / 2;
					writeRect.w = levelCompleteGraph->w;
					writeRect.h = levelCompleteGraph->h;
					SDL_BlitSurface(levelCompleteGraph, NULL, screen, &writeRect);
				}

				
				//Flip the surface
				SDL_Flip(screen);
			}
		}

		while (SDL_PollEvent(&event))
		{
			switch(event.type)
			{
			case SDL_KEYDOWN:
				// Allow Print Screen
				if ((event.key.keysym.sym == SDLK_PRINT) || (event.key.keysym.sym == SDLK_F10))
				{
					SDL_SaveBMP(screen, "SCREENSHOT.BMP");
					break;
				}
				else if (event.key.keysym.sym == SDLK_LEFT)
				{
					if (gameState == GAME_PLAY)
						heroSprite->moveLeft();
				}
				else if (event.key.keysym.sym == SDLK_RIGHT)
				{
					if (gameState == GAME_PLAY)
						heroSprite->moveRight();
				}
				else if (event.key.keysym.sym == SDLK_UP)
				{
					if (gameState == GAME_PLAY)
						heroSprite->moveUp();
				}
				else if (event.key.keysym.sym == SDLK_DOWN)
				{
					if (gameState == GAME_PLAY)
						heroSprite->moveDown();
				}
				else if ((event.key.keysym.sym == SDLK_q) ||
					(event.key.keysym.sym == SDLK_ESCAPE))
				{
					bQuit = true;
					break;
				}

				if (gameState == GAME_TITLE)
				{
					if ((event.key.keysym.sym != SDLK_PRINT) &&
						(event.key.keysym.sym != SDLK_F10) && 
						(event.key.keysym.sym != SDLK_q) && 
						(event.key.keysym.sym != SDLK_ESCAPE))
					gameState =  GAME_NEW_LEVEL;
				}

				else if (gameState == GAME_OVER)
				{
					if ((event.key.keysym.sym != SDLK_PRINT) &&
						(event.key.keysym.sym != SDLK_F10) && 
						(event.key.keysym.sym != SDLK_q) && 
						(event.key.keysym.sym != SDLK_ESCAPE))
					gameState =  GAME_TITLE;
				}
				else if (gameState == GAME_LEVEL_COMPLETE)
				{
					if ((event.key.keysym.sym != SDLK_PRINT) &&
						(event.key.keysym.sym != SDLK_F10) && 
						(event.key.keysym.sym != SDLK_q) && 
						(event.key.keysym.sym != SDLK_ESCAPE))
					gameState =  GAME_NEW_LEVEL;
				}
				break;
			case SDL_KEYUP:
				if (event.key.keysym.sym == SDLK_p)
				{
					if (gameState == GAME_PAUSE)
					{
						gameState = GAME_PLAY;
					}
					else if (gameState == GAME_PLAY)
					{
						gameState = GAME_PAUSE;
					}
				}
				if (gameState == GAME_PLAY)
					heroSprite->setKeyDown(false);
				break;
			case SDL_QUIT:
				bQuit = true;
				break;
			default:
				break;
			}
		}
	}

	//If playing sounds, stop so we can exit.
	if (Mix_Playing(-1)>0)
		Mix_HaltChannel(-1); 

	Mix_FreeChunk(welcomeSound);
	welcomeSound = NULL;

	Mix_FreeChunk(gameOverSound);
	gameOverSound = NULL;

	Mix_FreeChunk(levelCompleteSound);
	levelCompleteSound = NULL;

	SDL_FreeSurface(gameOverGraph);

	SDL_FreeSurface(levelCompleteGraph);

	SDL_FreeSurface(pauseGraph);

	SDL_FreeSurface(enemyGraph);

	SDL_FreeSurface(exitGraph);

	SDL_FreeSurface(heroGraph);
	
	SDL_FreeSurface(tiles);

	SDL_FreeSurface(titleGraph);

	for (i=0;i<MAX_ENEMIES;++i)
	{
		if (enemyList[i] != NULL)
		{
			delete enemyList[i];
		}
	}
	
	delete exitSprite;

	delete heroSprite;

	delete mapInfo;
	
	delete[] map;

	return(0);
}